from abstract_database import *
from abstract_apis import *
from .utils import *
import requests
async def get_rate_limit_url(method_name, *args, **kwargs):
    url = getEndpointUrl("/rate_limit")
    for arg in args:
        if "method" not in kwargs:
            kwargs["method"] = arg
    return await asyncGetRequest(url, kwargs)
    
async def log_response(method_name, response_data, *args, **kwargs):
    url = getEndpointUrl("/log_response")
    # Assuming that `args` is supposed to update `kwargs` for unspecified keys
    for arg in args:
        if "method" not in kwargs:
            kwargs["method"] = arg
        if "response_data" not in kwargs:
            kwargs["response_data"] = response_data
    # Use a direct payload if kwargs are not being used effectively
    payload = {
        "method": kwargs.get("method", method_name),
        "response_data": kwargs.get("response_data", response_data)
    }
    return await asyncPostRequest(url, kwargs, endpoint=endpoint)





from .utils import getEndpointUrl
from abstract_apis import asyncPostRequest
import asyncio
def callSolcatcherRpc(endpoint=None,**kwargs):
  url = getEndpointUrl(endpoint)
  return asyncio.run(asyncPostRequest(url=url,data=kwargs))
  



from abstract_solcatcher.utils import *
from abstract_apis import *

def callRequest(endpoint,*args,**kwargs):
  url = getEndpointUrl(endpoint)
  return postRequest(url,kwargs)

def getCallArgs(endpoint):
  return {'getMetaData': ['signature'], 'getPoolData': ['signature'], 'getTransactionData': ['signature'], 'getPoolInfo': ['signature'], 'getMarketInfo': ['signature'], 'getKeyInfo': ['signature'], 'getLpKeys': ['signature'], 'process': ['signature']}.get(get_endpoint(endpoint))




from .abstract_call import *
from .abstract_flask import *
from .abstract_async_call import *
from .abstract_async_flask import *
from .abstract_async_rate_limit import *
from .abstract_rate_limit import *
from .abstract_swap import *
from .async_abstract_call import *
from .utils import *
from .abstract_async_requests import *
from .abstract_solana.abstract_transactions import *
from .abstract_solana.address_lookup import *



from abstract_database import *
from abstract_apis import *
from .utils import *
def get_rate_limit_url(method_name,*args, **kwargs):
    url = getEndpointUrl("rate_limit")
    kwargs["method"]=method_name or kwargs.get("method")
    for arg in args:
        if "method" not in kwargs:
            kwargs["method"] = arg

    return getRequest(url, data=kwargs,response_result='url')
def log_response(method_name, response_data,*args,**kwargs):
    url = getEndpointUrl("log_response")
    payload = {
        "method": kwargs.get("method", method_name),
        "response_data": kwargs.get("response_data", response_data)
    }
    payload.update(kwargs) 
    for arg in args:
        if "method" not in payload:
            payload["method"] = arg
        if "response_data" not in payload:
            payload["response_data"] = arg
    return postRequest(url, data=payload)




from .utils import *
from abstract_apis import *

async def asyncCallRequest(endpoint,*args,**kwargs):
  endpoint = make_endpoint(endpoint)
  return await asyncPostRequest(getCallUrl(),kwargs,endpoint=endpoint)



from abstract_utilities import get_any_value
from .abstract_flask import viewTable
from .abstract_call import callRequest
class PoolManager:
    def __init__(self):
        self.poolInfo = viewTable('pool_info', latest=True)[0]
        self.update_tokens()

    def update_tokens(self):
        self.tokens = {}
        for typ in ['base', 'lp', 'quote']:
            self.tokens[typ] = self.get_token_info(typ)

    def get_token_info(self, typ='base'):
        typ = typ.lower()
        Mint = self.poolInfo.get(f'{typ}_mint')
        Reserve = get_any_value(self.poolInfo, f'{typ}_reserve')
        MetaData = callRequest('getMetaData', **{f'{typ}Mint': Mint})
        AccountInfo = callRequest('getAccountInfo', account=Mint)
        ProgramId = get_any_value(AccountInfo, 'owner')
        Decimals = get_any_value(AccountInfo, 'decimals')
        return {
            "decimals": Decimals,
            "programId": ProgramId,
            "metaData": MetaData,
            "reserve": Reserve,
            "mint": Mint
        }

    def calculate_amount_out(self, amount, fee=0.001):
        reserve_in = self.tokens['quote']['reserve']
        reserve_out = self.tokens['base']['reserve']
        decimals_in = self.tokens['quote']['decimals']
        decimals_out = self.tokens['base']['decimals']

        amount_in = amount * (10 ** decimals_in)
        fees = amount_in * fee
        amount_in_with_fee = amount_in - fees
        amount_out_raw = (reserve_out * amount_in_with_fee) / (reserve_in + amount_in_with_fee)

        return amount_out_raw / (10 ** decimals_out)

    def calculate_amount_in(self, amount, fee=0.001):
        reserve_in = self.tokens['base']['reserve']
        reserve_out = self.tokens['quote']['reserve']
        decimals_in = self.tokens['base']['decimals']
        decimals_out = self.tokens['quote']['decimals']

        amount_in = amount * (10 ** decimals_in)
        fees = amount_in * fee
        amount_in_with_fee = amount_in - fees
        amount_out_raw = (reserve_out * amount_in_with_fee) / (reserve_in + amount_in_with_fee)

        return amount_out_raw / (10 ** decimals_out)





from abstract_solcatcher import eatAll,make_list,callSolcatcherRpc
def get_txn_data(file_path=None):
  file_path = file_path or 'txnData.json'
  return safe_read_from_json(file_path)
def get_account_keys(txnData):
    return txnData['transaction']['message']['accountKeys']

def get_loaded_addresses(txnData):
    return txnData['meta']['loadedAddresses']

def get_read_only_addresses(txnData):
    return get_loaded_addresses(txnData).get('readonly', [])

def get_writable_addresses(txnData):
    return get_loaded_addresses(txnData).get('writable', [])

def get_all_account_keys(txnData):
  accountKeys = get_account_keys(txnData) 
  accountKeys+= get_read_only_addresses(txnData)
  accountKeys+= get_writable_addresses(txnData)
  return accountKeys

class saveAcountKeys:
  def __init__(self,txnData):
    self.accountKeys = get_account_keys(txnData)
    self.loaded_addresses = get_loaded_addresses(txnData)
    self.read_only_addresses = get_read_only_addresses(txnData)
    self.writable_addresses = get_writable_addresses(txnData)
    self.accountKeys = self.get_all_addresses()
    
  def get_all_addresses(self):
    self.accountKeys = self.accountKeys
    self.accountKeys += self.read_only_addresses
    self.accountKeys += self.writable_addresses
    return self.accountKeys

  def get_account_from_index(self,index):
    account = None
    if len(self.accountKeys)>index:
      account = self.accountKeys[index]
    return account

  def get_account_keys_from_instruction(self,instruction):
    indexes = instruction.get('accounts')
    accounts = []
    for index in indexes:
      accounts.append(self.get_account_from_index(index))
    return accounts
  def add_accounts_to_instruction(self,instruction):
    accounts = self.get_account_keys_from_instruction(instruction)
    instruction['extractedAccounts'] = accounts
    return instruction
  def add_accounts_to_instructions(self,instructions):
    for i,instruction in enumerate(instructions):
      instruction = self.add_accounts_to_instruction(instruction)
      instructions[i] = instruction
    return instructions

  
  def get_programIdIndex_from_instruction(self,instruction):
    index = instruction.get('programIdIndex')
    programId = self.get_account_from_index(index)
    return programId
  def add_programId_to_instruction(self,instruction):
    programId = self.get_programIdIndex_from_instruction(instruction)
    instruction['programId'] = programId
    return instruction
  def add_programIds_to_instructions(self,instructions):
    for i,instruction in enumerate(instructions):
      instruction = self.add_programId_to_instruction(instruction)
      instructions[i] = instruction
    return instructions

def get_log_messages(txnData):
    return txnData['meta']['logMessages']

class logMessageManager:
      def __init__(self,txnData):
        self.logMessages = get_log_messages(txnData)
        self.logClusters = self.get_log_index_clusters()
      def get_log_index_clusters(self):
        clusterCount = 0
        self.logClusters =[]
        for i,log in enumerate(self.logMessages):
          logLower = log.lower()
          if 'invoke' in logLower:
            self.logClusters.append({"logs":[],"clusterIndex":clusterCount,"type":None,"types":[],"start":i,"end":i,"programId":log.split(' ')[1]})
          if 'success' in logLower:
            self.logClusters[-1]["end"]=i
          self.logClusters = self.get_log_type(log)
        return self.logClusters
      def get_log_type(self,log):
        currentLogClusterJs = {}
        logSpl = [eatAll(logSp,[' ',':']) for logSp in log.split(' ') if logSp][1:]
        currentLogClusterJs = {"programId":logSpl[0],"type":logSpl[1]}
        for i,log in enumerate(logSpl):
          if log == 'log':
            currentLogClusterJs['type'] = logSpl[-1] or logSpl[-2]
            self.logClusters[-1]["type"] = currentLogClusterJs['type']
        if len(logSpl)>2:
         currentLogClusterJs["action"]=logSpl[2]
        if len(logSpl)>2:
         currentLogClusterJs["vars"]=' '.join(logSpl[3:])
        self.logClusters[-1]["types"].append(currentLogClusterJs['type'])
        self.logClusters[-1]["logs"].append(currentLogClusterJs)
        return self.logClusters
      def get_cluster_type_from_index(self,index):
        cluster = {}
        if len(self.logClusters) >index:
          cluster = self.logClusters[index]
        clusterType = make_list(cluster.get("type") or cluster.get('types'))
        if clusterType:
          clusterType = clusterType[0]
        return clusterType
      def add_cluster_type_to_instruction(self,index,instruction):
        clusterType = self.get_cluster_type_from_index(index)
        instruction['type']=clusterType
        return instruction
      
      def add_cluster_types_to_instructions(self,instructions,indexStart=0):
        new_instructions = []
        for i,instruction in enumerate(instructions):
          index = i+indexStart
          instruction = self.add_cluster_type_to_instruction(index,instruction)
          new_instructions.append(instruction)
        return new_instructions
  
def get_instructions(txnData):
  return txnData['transaction']['message']['instructions']

def get_inner_instructions(txnData):
  return txnData['meta']['innerInstructions'][0]['instructions']

def update_instructions(txnData,instructions):
  txnData['transaction']['message']['instructions'] = instructions
  return  txnData
def update_inner_instructions(txnData,inner_instructions):
  txnData['meta']['innerInstructions'][0]['instructions'] = inner_instructions
  return txnData

class saveInstructions:
  def __init__(self,txnData,account_mgr=None,log_mgr=None):
    self.account_mgr = account_mgr or saveAcountKeys(txnData)
    self.log_mgr = log_mgr or logMessageManager(txnData)
    self.instructions = get_instructions(txnData)
    self.inner_instructions = get_inner_instructions(txnData)
    self.allInstructions = self.make_all_associations()
    
  def associate_accounts_with_instructions(self):
    self.instructions = self.account_mgr.add_accounts_to_instructions(self.instructions)
    return self.instructions

  def associate_accounts_with_inner_instructions(self):
    self.inner_instructions = self.account_mgr.add_accounts_to_instructions(self.inner_instructions)
    return self.inner_instructions

  def associate_programIdIndexs_with_instructions(self):
    self.instructions = self.account_mgr.add_programIds_to_instructions(self.instructions)
    return self.inner_instructions

  def associate_programIdIndexs_with_inner_instructions(self):
    self.inner_instructions = self.account_mgr.add_programIds_to_instructions(self.inner_instructions)
    return self.inner_instructions

  def associate_types_with_instructions(self):
    self.instructions = self.log_mgr.add_cluster_types_to_instructions(self.instructions)
    return self.instructions

  def associate_types_with_inner_instructions(self):
    index_start = len(self.instructions)
    self.inner_instructions = self.log_mgr.add_cluster_types_to_instructions(self.inner_instructions,indexStart=index_start)
    return self.inner_instructions

  def associate_all_instructions(self):
    self.instructions = self.associate_types_with_instructions()
    self.instructions = self.associate_accounts_with_instructions()
    self.instructions = self.associate_programIdIndexs_with_instructions()
    return self.instructions
  
  def associate_all_inner_instructions(self):
    self.inner_instructions = self.associate_types_with_inner_instructions()
    self.inner_instructions = self.associate_accounts_with_inner_instructions()
    self.inner_instructions = self.associate_programIdIndexs_with_inner_instructions()
    return self.inner_instructions

  def get_all_instructions(self):
    self.allInstructions = self.instructions
    self.allInstructions += self.inner_instructions
    return self.allInstructions

  def make_all_associations(self):
    self.instructions = self.associate_all_instructions()
    self.inner_instructions = self.associate_all_inner_instructions()
    return self.get_all_instructions()

class TransactionProcessor:
    def __init__(self, txnData=None,signature=None,instruction_mgr = None,account_mgr=None,log_mgr=None,**kwargs):
        txnData = txnData or self.getTransacion(signature,**kwargs)
        self.txnData = txnData
        self.log_mgr = log_mgr or logMessageManager(txnData)
        self.account_mgr = account_mgr or saveAcountKeys(txnData)
        self.instruction_mgr = instruction_mgr or saveInstructions(self.txnData,log_mgr=self.log_mgr,account_mgr=self.account_mgr)
    def get_instructions(self):
      return self.instruction_mgr.allInstructions

    def get_log_clusters(self):
      return self.log_mgr.logClusters

    def get_account_keys(self):
      return self.account_mgr.accountKeys

    def get_txn_data(self):
      input(self.txnData)
      txnData['transaction']['message']['instructions'] = self.instruction_mgr.instructions
      txnData['meta']['innerInstructions'][0]['instructions'] = self.instruction_mgr.inner_instructions
      return self.txnData
    
    def getTransacion(self,signature,**kwargs):
      kwargs['signature']=signature
      return callSolcatcherRpc('getTransaction',**kwargs)
def get_transaction(signature,**kwargs):
  kwargs['signature']=signature
  txn_mgr = TransactionProcessor(**kwargs)
  return txn_mgr.get_txn_data()




from abstract_solcatcher.abstract_async_requests import callSolcatcherRpc
from abstract_utilities import eatAll,make_list,get_any_value
from abstract_bots.bot_functions import divide_it,return_0,exponential
from abstract_bots.txn_functions import get_any_value,get_account_keys
import pandas as pd
def search_for_account_index(data,index_number):
    for index_data in data:
        if str(index_data.get('accountIndex')) == str(index_number):
            return index_data
          
def get_txn_data(file_path=None):
  file_path = file_path or 'txnData.json'
  return safe_read_from_json(file_path)
def get_account_keys(txnData):
    return txnData['transaction']['message']['accountKeys']

def get_loaded_addresses(txnData):
    return txnData['meta']['loadedAddresses']

def get_read_only_addresses(txnData):
    return get_loaded_addresses(txnData).get('readonly', [])

def get_writable_addresses(txnData):
    return get_loaded_addresses(txnData).get('writable', [])

def get_all_account_keys(txnData):
  accountKeys = get_read_only_addresses(txnData)
  accountKeys += get_writable_addresses(txnData)
  accountKeys += get_account_keys(txnData) 

  return accountKeys

class saveAcountKeys:

  def __init__(self,txnData):
    self.accountKeys = get_account_keys(txnData)
    self.loaded_addresses = get_loaded_addresses(txnData)
    self.read_only_addresses = get_read_only_addresses(txnData)
    self.writable_addresses = get_writable_addresses(txnData)
    self.accountKeys = self.get_all_addresses()
  def get_all_addresses(self):
    all_keys = self.accountKeys + self.read_only_addresses + self.writable_addresses
    return self.sort_account_keys(all_keys)
    
  def sort_account_keys(self, accountKeys):
    return sorted(accountKeys, key=lambda x: x['pubkey'].to_base58())    
  def get_all_addresses(self):
    self.accountKeys = self.accountKeys
    self.accountKeys += self.read_only_addresses
    self.accountKeys += self.writable_addresses
    return self.accountKeys

  def get_account_from_index(self,index):
    account = None
    if len(self.accountKeys)>index:
      account = self.accountKeys[index]
    return account

  def get_account_keys_from_instruction(self,instruction):
    indexes = instruction.get('accounts')
    accounts = []
    for index in indexes:
      accounts.append(self.get_account_from_index(index))
    return accounts
  def add_accounts_to_instruction(self,instruction):
    accounts = self.get_account_keys_from_instruction(instruction)
    instruction['extractedAccounts'] = accounts
    return instruction
  def add_accounts_to_instructions(self,instructions):
    for i,instruction in enumerate(instructions):
      instruction = self.add_accounts_to_instruction(instruction)
      instructions[i] = instruction
    return instructions

  
  def get_programIdIndex_from_instruction(self,instruction):
    index = instruction.get('programIdIndex')
    programId = self.get_account_from_index(index)
    return programId
  def add_programId_to_instruction(self,instruction):
    programId = self.get_programIdIndex_from_instruction(instruction)
    instruction['programId'] = programId
    return instruction
  def add_programIds_to_instructions(self,instructions):
    for i,instruction in enumerate(instructions):
      instruction = self.add_programId_to_instruction(instruction)
      instructions[i] = instruction
    return instructions

def get_log_messages(txnData):
    return txnData['meta']['logMessages']

class logMessageManager:
      def __init__(self,txnData):
        self.logMessages = get_log_messages(txnData)
        self.logClusters = self.get_log_index_clusters()
      def get_log_index_clusters(self):
        clusters =[]
        clusterCount=-1
        for j,log in enumerate(self.logMessages):
            curr_logs = {}
            
            # Split the log into parts based on whitespace
            parts = [part.replace(':','') for part in log.split()]
            
            if not parts:
                continue  # Skip empty lines
            # Parse based on the structure
            if "invoke" in parts:
                
                clusterCount+=1
                clusters.append({"logs":[],"clusterIndex":clusterCount,"programId":parts[1],"stackHeight":str(parts[-1])[1:-1],"start":j,"end":j,"types":[]})
                
                curr_logs['cmdType'] = parts[0]
                curr_logs['programId'] = parts[1]
                curr_logs['event'] = parts[2]  # 'invoke' or 'success'
                curr_logs['vars'] = parts[3] if len(parts) > 3 else None
            
            elif "consumed" in parts:
                curr_logs['cmdType'] = parts[0]
                curr_logs['programId'] = parts[1]
                curr_logs['event'] = 'consumed'
                consumed_index = parts.index('consumed')
                curr_logs['vars'] = " ".join(parts[consumed_index + 1:])
            
            elif "return:" in log:
                curr_logs['cmdType'] = parts[0]
                curr_logs['programId'] = parts[1]
                curr_logs['event'] = 'return'
                curr_logs['vars'] = " ".join(parts[2:])
            
            elif "log:" in log:
                curr_logs['cmdType'] = parts[0]
                curr_logs['programId'] = parts[1]
                log_index = parts.index('log')
                curr_logs['event'] = parts[log_index + 1]
                curr_logs['type'] = parts[log_index + 2] if len(parts) > log_index + 2 else None
                clusters[-1]['types'].append(curr_logs['type'])
                if clusters[-1].get('type') == None:
                  clusters[-1]['type']=curr_logs['type']
                curr_logs['vars'] = " ".join(parts[log_index + 3:]) if len(parts) > log_index + 3 else None
            
            else:
                curr_logs['cmdType'] = parts[0]
                curr_logs['programId'] = parts[1]
                curr_logs['event'] = parts[2] if len(parts) > 2 else None
                curr_logs['vars'] = " ".join(parts[3:]) if len(parts) > 3 else None
            clusters[-1]['end']=j
            # Store parsed log in the list
            clusters[-1]["logs"].append(curr_logs)
        return clusters      
      def get_cluster_type_from_index(self,index):
        cluster = {}
        if len(self.logClusters) >index:
          cluster = self.logClusters[index]
        clusterType = make_list(cluster.get("type") or cluster.get('types'))
        if clusterType:
          clusterType = clusterType[0]
        return clusterType
      def add_cluster_type_to_instruction(self,index,instruction):
        clusterType = self.get_cluster_type_from_index(index)
        instruction['type']=clusterType
        return instruction
      
      def add_cluster_types_to_instructions(self,instructions,indexStart=0):
        new_instructions = []
        for i,instruction in enumerate(instructions):
          index = i+indexStart
          instruction = self.add_cluster_type_to_instruction(index,instruction)
          new_instructions.append(instruction)
        return new_instructions
  
def get_instructions(txnData):
  return txnData['transaction']['message']['instructions']

def get_inner_instructions(txnData):
  return txnData['meta']['innerInstructions'][0]['instructions']

def update_instructions(txnData,instructions):
  txnData['transaction']['message']['instructions'] = instructions
  return  txnData
def update_inner_instructions(txnData,inner_instructions):
  txnData['meta']['innerInstructions'][0]['instructions'] = inner_instructions
  return txnData

class saveInstructions:
  def __init__(self,txnData,account_mgr=None,log_mgr=None):
    self.account_mgr = account_mgr or saveAcountKeys(txnData)
    self.log_mgr = log_mgr or logMessageManager(txnData)
    self.instructions = get_instructions(txnData)
    self.inner_instructions = get_inner_instructions(txnData)
    self.allInstructions = self.make_all_associations()
    
  def associate_accounts_with_instructions(self):
    self.instructions = self.account_mgr.add_accounts_to_instructions(self.instructions)
    return self.instructions

  def associate_accounts_with_inner_instructions(self):
    self.inner_instructions = self.account_mgr.add_accounts_to_instructions(self.inner_instructions)
    return self.inner_instructions

  def associate_programIdIndexs_with_instructions(self):
    self.instructions = self.account_mgr.add_programIds_to_instructions(self.instructions)
    return self.inner_instructions

  def associate_programIdIndexs_with_inner_instructions(self):
    self.inner_instructions = self.account_mgr.add_programIds_to_instructions(self.inner_instructions)
    return self.inner_instructions

  def associate_types_with_instructions(self):
    self.instructions = self.log_mgr.add_cluster_types_to_instructions(self.instructions)
    return self.instructions

  def associate_types_with_inner_instructions(self):
    index_start = len(self.instructions)
    self.inner_instructions = self.log_mgr.add_cluster_types_to_instructions(self.inner_instructions,indexStart=index_start)
    return self.inner_instructions

  def associate_all_instructions(self):
    self.instructions = self.associate_types_with_instructions()
    self.instructions = self.associate_accounts_with_instructions()
    self.instructions = self.associate_programIdIndexs_with_instructions()
    return self.instructions
  
  def associate_all_inner_instructions(self):
    self.inner_instructions = self.associate_types_with_inner_instructions()
    self.inner_instructions = self.associate_accounts_with_inner_instructions()
    self.inner_instructions = self.associate_programIdIndexs_with_inner_instructions()
    return self.inner_instructions

  def get_all_instructions(self):
    self.allInstructions = self.instructions
    self.allInstructions += self.inner_instructions
    return self.allInstructions

  def make_all_associations(self):
    self.instructions = self.associate_all_instructions()
    self.inner_instructions = self.associate_all_inner_instructions()
    return self.get_all_instructions()


class TransactionProcessor:
    def __init__(self, txnData=None,signature=None,instruction_mgr = None,account_mgr=None,log_mgr=None,**kwargs):
        txnData = txnData or self.getTransacion(signature,**kwargs)
        self.txnData = txnData
        self.log_mgr = log_mgr or logMessageManager(txnData)
        self.account_mgr = account_mgr or saveAcountKeys(txnData)
        self.instruction_mgr = instruction_mgr or saveInstructions(self.txnData,log_mgr=self.log_mgr,account_mgr=self.account_mgr)
    def get_instructions(self):
      return self.instruction_mgr.allInstructions

    def get_log_clusters(self):
      return self.log_mgr.logClusters

    def get_account_keys(self):
      return self.account_mgr.accountKeys

    def get_txn_data(self):
      self.txnData['transaction']['message']['instructions'] = self.instruction_mgr.instructions
      self.txnData['meta']['innerInstructions'][0]['instructions'] = self.instruction_mgr.inner_instructions
      return self.txnData
    
    def getTransacion(self,signature,**kwargs):
      kwargs['signature']=signature
      return callSolcatcherRpc('getTransaction',**kwargs)

def get_test_signature():
  return "RwJcLtyVyPSBWfNCzNcCaqKrmTmaKqDezzMFXTubmPGM9AfjdMW2tg9Xb1Dm8ZvHDYsH3Za22zWq2N8sQKxd1uq"
def get_txnMgr(txnData=None,signature=None):
  return TransactionProcessor(txnData=txnData,signature=signature)
class txnManager:
    def __init__(self, txnData=None, signature=None, txn_mgr=None, acct_mgr=None):
        self.signature = signature or get_test_signature()
        self.txn_mgr = txn_mgr or get_txnMgr(txnData=txnData, signature=self.signature)
        self.acct_mgr = acct_mgr or self.txn_mgr.account_mgr
        self.txnData = self.txn_mgr.txnData or get_transaction(self.signature)
        self.preBalances = self.txnData['meta']['preBalances']
        self.postBalances = self.txnData['meta']['postBalances']
        self.preTokenBalances = self.txnData['meta']['preTokenBalances']
        self.postTokenBalances = self.txnData['meta']['postTokenBalances']
        self.all_txns = {"Address":[],"Owner":[],'Balance Before':[],"Balance After":[],"Change":[],"Token":[]}
        self.accountKeys = self.acct_mgr.accountKeys
        self.all_txns = {"Address": [], "Owner": [], 'Balance Before': [], "Balance After": [], "Change": [], "Token": []}
        
        # Create a dictionary for quick lookups by accountIndex
        self.index_to_postToken = {token['accountIndex']: token for token in self.postTokenBalances}
    def get_from_postToken_index(self,index):
      return search_for_account_index(self.postTokenBalances,index)
    def get_postToken_amnt_dict(self,index):
      return self.get_from_postToken_index(index)
    def get_from_preToken_index(self,index):
      return search_for_account_index(self.preTokenBalances,index)
    def get_preToken_amnt_dict(self,index):
      return self.get_from_preToken_index(index)
    def get_amount_dict(self, amount, decimals=9):
        if isinstance(amount, dict):
            amount_dict = get_any_value(amount, 'uiTokenAmount')
            amount = get_any_value(amount_dict, 'amount')
            decimals = get_any_value(amount_dict, 'decimals')
        return exponential(amount, decimals, -1)

    def create_token_txns(self):
        dont_use = []
        for i,balances in enumerate([self.preTokenBalances,self.postTokenBalances]):
            for post in balances:
                index = post.get('accountIndex')
                if index not in dont_use:
                    dont_use.append(index)
                    after = self.get_amount_dict(self.get_from_postToken_index(index))
                    change = self.get_amount_dict(post)
                    if after!=None and change !=None:
                        before = after-change
                        address = self.acct_mgr.get_account_from_index(index)
                        owner = post.get('owner')
                        token = post.get('mint')
                        if i == 0:
                            pre_change = change
                            change = before
                            before = pre_change
                        self.all_txns["Address"].append(address)
                        self.all_txns["Owner"].append(owner)
                        self.all_txns['Balance Before'].append(before)
                        self.all_txns["Balance After"].append(after)
                        self.all_txns["Change"].append(change)
                        self.all_txns["Token"].append(token)
        return self.all_txns
      
    def visualize_txns(self):
        if not self.all_txns['Address']:  # Check if transactions are populated
            self.create_token_txns()  # Populate the transactions if not already done
        
        df = pd.DataFrame(self.all_txns)
        pd.set_option('display.max_columns', None)  # To display all columns
        print(df)

        return df
      
def get_token_txn_data(txnData=None,signature=None):
  txn_mgr = txnManager(txnData=txnData,signature=signature)
  return txn_mgr.create_token_txns()

def display_txn_data(txnData=None,signature=None):
  txn_mgr = txnManager(txnData=txnData,signature=signature)
  return txn_mgr.visualize_txns()

def get_transaction(signature,**kwargs):
  kwargs['signature']=signature
  txnData_mgr = TransactionProcessor(**kwargs)
  return txnData_mgr.get_txn_data()




from .utils import *
from abstract_apis import *

async def asyncCallRequest(endpoint,*args,**kwargs):
  endpoint = make_endpoint(endpoint)
  return await asyncPostRequest(getCallUrl(),kwargs,endpoint=endpoint)

def getCallArgs(endpoint):
  return {'getMetaData': ['signature'], 'getPoolData': ['signature'], 'getTransactionData': ['signature'], 'getPoolInfo': ['signature'], 'getMarketInfo': ['signature'], 'getKeyInfo': ['signature'], 'getLpKeys': ['signature'], 'process': ['signature']}.get(get_endpoint(endpoint))




from abstract_apis import get_url,make_endpoint
def getSolcatcherUrl():
  return 'https://solcatcher.io'
def getEndpointUrl(endpoint=None,url=None):
  url = url or getSolcatcherUrl()
  endpoint = make_endpoint(endpoint or '/')
  return get_url(url,endpoint)
def updateData(data,**kwargs):
  data.update(kwargs)
  return data



from .utils import *
from .abstract_call import *
from abstract_apis import *
def postFlaskRequest(endpoint,**kwargs):
  url = getEndpointUrl(endpoint)
  return postRequest(url,kwargs)

def getFlaskRequest(endpoint,**kwargs):
  url = getEndpointUrl(endpoint)
  return getRequest(url,kwargs)

def viewTable(table_name, column_name=None, start=None, end=None, filters=None, search_string=None,deep_search=False,latest=None,**kwargs):
  return postFlaskRequest('view_table',table_name=table_name,column_name=column_name, start=start, end=end, filters=filters, search_string=search_string,deep_search=deep_search,latest=latest,**kwargs)

def getLpKeys(table_name=None, column_name=None, start=None, end=None, filters=None, search_string=None,deep_search=False,latest=None,**kwargs):
  response = viewTable(table_name=table_name or 'key_info',column_name=column_name, start=start, end=end, filters=filters, search_string=search_string,deep_search=deep_search,latest=latest,**kwargs)
  signature= response[0].get('signature')
  return callRequest('getLpKeys',signature)

def list_tables(with_data=True):
  url = getEndpointUrl(endpoint)
  response = getRequest(url, data={"with_data":with_data})
  return response

def list_columns(table_name=None):
  tables = {}
  url = getEndpointUrl(endpoint)
  table_names = make_list(table_name or list_tables(with_data=True))
  for tableName in table_names:
    tables[tableName]= getRequest(url, data={"table_name":tableName})
  return tables

